#include <algorithm> // 'std::transform'...
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <memory> // 'std::unique_ptr'...
#include <string> 
#include <type_traits> // 'std::is_same'...
#include <unordered_map>
#include <utility> // 'std::forward'...
#include <vector>

#include "Smart_pointers.h"

void Elephantidae::push_family_member(const std::shared_ptr<Elephantidae>& pe) {
	_family_members.push_back(pe);
}

std::shared_ptr<Elephantidae> Elephantidae::peek_family_member() const {
	return _family_members.back();
}

bool Elephant::operator<(const Elephant& e) const {
	return this->_weight < e._weight;
}

void* Elephant::_Preallocated_memory = nullptr;

void* Elephant::operator new(size_t size) noexcept(false) {
	if (_Preallocated_memory) return _Preallocated_memory;
	void* p = std::malloc(size);
	// Alternative (caution: if 'new' is *inherited* then the allocated size is wrong):
	// void* p = ::new Elephant(); // Call of global 'new' to avoid recursion...
	// Solutions: (1) use 'final' for blocking inheritance from 'Elephant', or (2) use
	// private inheritance, i.e., 'class African_elephant : private Elephant',
	// which prevents undesired 'new' inheritance: 'African_elephant* ae = new African_elephant;' -> compilation error
	if (!p) throw std::bad_alloc(); // <=> 'p == nullptr' and 'ptr == NULL'
	return p;
}

void Elephant::operator delete(void* p) noexcept(true) {
	if (_Preallocated_memory) std::free(_Preallocated_memory);
	// Alternative:
	// 'if (_Preallocated_memory) ::delete _Preallocated_memory;' // Call of global 'delete' to avoid recursion...
	_Preallocated_memory = p; // Recycle...
}

std::string Elephant::sub_species() const {
	if (std::is_same<decltype(this), const African_elephant* const>::value) return _Loxodonta_africana;
	if (std::is_same<decltype(this), const Asian_elephant* const>::value) return _Elephas_maximus;
	return "?";
}
const std::string Elephant::_Loxodonta_africana = "Loxodonta africana";
const std::string Elephant::_Elephas_maximus = "Elephas maximus";

int main(int argc, char** argv) {
	std::cout << "Smart pointers..." << std::endl;

	Elephant* Babar = new Elephant;
	delete Babar;
	Elephant* Elmer = new Elephant;
	delete Elmer;
	assert(Babar == Elmer);

	// Modern C++ allows the allocation of pointers without the need of user-defined deletions:
	std::unique_ptr<Mammoth> m0(new Mammoth("m0")); // '#include <memory>'
	// std::unique_ptr<Mammoth> bug(m0.get()); // Hell and damnation!This may work only if 'm0.release();' occurs before resource destruction...
	std::unique_ptr<Mammoth> m1 = std::make_unique<Mammoth>("m1");
	// std::unique_ptr<Mammoth> m2 = m1; // Compilation error...
	// Mammoth m = *m1; // '*' operator overridden in 'std::unique_ptr' -> compilation error with VisualStudio 2019?
	Mammoth* pm = m1.get(); // Get the managed object, caution: double pointing!
	assert(pm != nullptr);
	pm = m1.release();
	assert(m1.get() == nullptr); // Single pointing, i.e., 'pm'

	std::unique_ptr<Mammoth> La_Madeleine(new Mammoth("Le mammouth gravé sur ivoire de La Madeleine (Tursac, Dordogne)"));
	std::unique_ptr<Mammoth>&& La_Madeleine_is_moved = std::move(La_Madeleine);
	//    std::cout << "La_Madeleine.get(): " << La_Madeleine.get() << std::endl;
	//    std::cout << "La_Madeleine_is_moved.get(): " << La_Madeleine_is_moved.get() << std::endl;
	assert(La_Madeleine.get() != nullptr && La_Madeleine.get() == La_Madeleine_is_moved.get());
	La_Madeleine.release();
	assert(La_Madeleine.get() == nullptr && La_Madeleine_is_moved.get() == nullptr);

	std::unique_ptr<Mammoth> La_Grande_Liakhov(new Mammoth("La Grande Liakhov, l'île aux mammouths"));
	Mammoth* La_Grande_Liakhov_resource = La_Grande_Liakhov.get();
	std::unique_ptr<Mammoth> La_Grande_Liakhov_is_moved = std::move(La_Grande_Liakhov);
	assert(La_Grande_Liakhov.get() == nullptr && La_Grande_Liakhov_resource == La_Grande_Liakhov_is_moved.get());

	std::shared_ptr<Mammoth> ma;
	{
		std::shared_ptr<Mammoth> mb = std::make_shared<Mammoth>();
		ma = mb; // In essence, no compilation error...
		assert(ma.use_count() == 2);
	} // No delete on 'mb' since shared with 'ma'
	assert(ma.use_count() == 1);

	std::shared_ptr<Elephantidae> Jumbo = std::make_shared<Elephant>();
	// 'Jumbo' is passed by reference (note that passing by copy would increment the number of pointers returned by 'use_count'):
	ma->push_family_member(Jumbo); // Overridden '->' operator in 'std::shared_ptr'
	// std::shared_ptr<Elephant> pe = ma->peek_family_member(); // Compilation error...
	std::shared_ptr<Elephant> pe = std::dynamic_pointer_cast<Elephant>(ma->peek_family_member());
	assert(pe);

	std::weak_ptr<Mammoth> mx; // Weak pointer aims at replacing "raw" pointer obtained by 'get'...
	{
		std::shared_ptr<Mammoth> my = std::make_shared<Mammoth>();
		mx = my;
		assert(my.use_count() == 1); // Weak pointer has no impact on the number of pointers returned by 'use_count'
		std::shared_ptr<Mammoth> mz = mx.lock(); // 'lock' copies into 'shared_ptr' object before usage...
		std::cout << "mz.get(): " << mz.get() << std::endl;
	} // Delete on 'my' and 'mz'
	assert(mx.use_count() == 0);

	std::unique_ptr<Mammoth> the_very_first_Mammoth(new Mammoth("the_very_first_Mammoth"));
	// Compilation error (capture by copy cannot apply on movable-only objects as 'std::unique_ptr' objects):
	/*auto my_lambda_expression = [the_very_first_Mammoth](std::unique_ptr<Mammoth>&& m) {
		m->best_friend(the_very_first_Mammoth.get());
	};*/
	// Capture with initializer (C++14):
	auto my_lambda_expression = [best_friend = the_very_first_Mammoth.get()](std::unique_ptr<Mammoth>&& m) {
		m->best_friend(best_friend);
	};
	my_lambda_expression(std::move(m0)); // 'the_very_first_Mammoth' becomes best friend of 'm0'...
	assert(m0.get() != nullptr); // Forced 'move' does not empty 'm0'...

	//    std::unique_ptr<Mammoth> zoo[] = {m0, m1}; // Cannot copy...
	std::unordered_map<std::string, std::unique_ptr<Mammoth>&& > zoo; // Indexed access based on names...
	std::cout << std::boolalpha <<
		std::is_same_v<decltype(zoo)::value_type, std::pair<const std::string, std::unique_ptr<Mammoth>&&>>
		<< std::endl; // 'true'

	std::unique_ptr<Mammoth> m = std::make_unique<Mammoth>("m");
	zoo.emplace(m.get()->name, std::move(m)); // 'zoo.try_emplace(m.get()->name, std::move(m));' // C++17
	assert(m.get() != nullptr); // Forced 'move' does not empty 'm'...

	decltype(zoo)::iterator i = zoo.begin(); // So, '*i' is a pair...
	auto&& value = (*i).second; // Type of 'value' is 'std::unique_ptr<Mammoth>&' and *NOT* 'std::unique_ptr<Mammoth>&&':
	static_assert(std::is_same_v<decltype(value), std::unique_ptr<Mammoth>&>, "Type checking failed, why?");

	//std::vector<std::unique_ptr<Mammoth> > values; // Arrrrrgggggllll: no 'values' function on maps!
	//std::transform(std::begin(zoo), std::end(zoo), std::back_inserter(values),
	//	[](/* auto */ /* -> from C++14 */ const decltype(zoo)::value_type& pair) {
	//		static_assert(std::is_same_v<decltype(pair.second), std::unique_ptr<Mammoth>&&>, "Type checking failed, why?");
	//		return std::move(pair.second); // This empties the original map!
	//	});
	//assert(m.get() == nullptr); // Forced 'move' emptied 'm'...

	std::vector<const Mammoth*> values;
	std::transform(std::begin(zoo), std::end(zoo), std::back_inserter(values),
		[](const auto& pair) {
			static_assert(std::is_same_v<decltype(pair.second), std::unique_ptr<Mammoth>&&>, "Type checking failed, why?");
			return pair.second.get();
		});
	assert(m.get() != nullptr);

	// Before terminating, please release pool, if any, pointed by '_Preallocated_memory'

	return 0;
}